/**************************************************************************************************
* NOTICE OF COPYRIGHT AND OWNERSHIP OF SOFTWARE:
*
* Copyright 2010 Flexera Software, Inc. and/or InstallShield Co. Inc.  All Rights Reserved.  
*
* This computer program is the property of Flexera Software, Inc. and InstallShield Co. Inc. of 
* Schaumburg, Illinois, U.S.A.  Any use, copy, publication, distribution, display, modification, 
* or transmission of this computer program in whole or in part in any form or by any means without
* the prior express written permission of Flexera Software, Inc. or InstallShield Co. Inc. is 
* strictly prohibited.
* 
* Except when expressly provided by Flexera Software, Inc. or InstallShield Co. Inc. in writing,
* possession of this computer program shall not be construed to confer any license or rights under 
* any of Flexera Software, Inc.s or InstallShield Co. Inc.s intellectual property rights, whether
* by estoppel, implication, or otherwise.  
* 
* ALL COPIES OF THIS PROGRAM MUST DISPLAY THIS NOTICE OF COPYRIGHT AND OWNERSHIP IN FULL.
**************************************************************************************************/
/*
	This sample utility demonstrates FNP Composite Transactions.

	It is intended as a resource for developers, validators and documenters of 
	FNP and FNP enabled products.  It is not intended for release to end-users, 
	rather as a model for the development of activation utilites for FNP enabled
	products.

	IN PARTICULAR the functionality it provides for cancelling stored requests 
	could be used by fraudulent users to obtain additional licences using a manual
	(offline) return transaction - by cancelling the request instead of processing
	the response.  This ability may need to be provided so that bona fide users 
	who start a return transaction but cannot complete can continue to check out 
	from the fulfillment.  The business rules must define the criteria for 
	allowing return requests to be cancelled.

	It is a command line utility written in 'C' to provide maximum coverage of 
	supported FNP platforms.  It is built and released as part of the kit in both
	source and binary form; it is re-built by the kit make file (makefile.act).
	It has extensive command line options, providing the ability to create 
	requests with all the elements define by the XML schema.

	The makefile builds the executable comptranutil and then uses preptool to
	produce two separate executables, one for use with application trusted 
	storage and one for server trusted storage (appcomptranutil and 
	servercomptranutil respectively).

	It uses the API defined by header FlxActTransaction.h to create composite
	requests and process their responses.  It can also manage outstanding 
	requests in trusted storage.
	
	As a convenience it can also view and delete fulfillment records.

	Source files

		compTranUtil.c					Main entry (this file)
		compTranUtilArgParser.c			Parse command line arguments to internal objects.
		compTranUtilCommand.c			Command support (get value, attributes, dictionary etc.)
		compTranUtilCommandDef.c		Definition of command line (includes -help)
		compTranUtilContext.c			Context (command list, options, status, handles etc.)
		compTranUtilCreateRequest.c		Composite request creation (-new -activate -return -repair)
		compTranUtilFulfillments.c		Fulfillment Record view and delete (-view, -deleteFR).
		compTranUtilHelper.c			Helper functions (memory, string, print) 
		compTranUtilManageRequests.c	Manage stored requests (-listRequests, -cancelRequests)
		compTranUtilResponse.c			Process responses (-process)
		compTranUtilSend.c				Send response to server and process response (-transaction)

	Function naming conventions

		flxActTranXXX		Composite transaction API functions defined in FlxActTransaction.h 
		flxActCommonXXX		Other Activation API functions defined in FlxActCommon.h 
		flxCtuXXX			Functions within this utility.

	Usage

		Use the -help command to print a usage summary; it defines the other help options, e.g.
			appcomptranutil -help
			servercomptranutil -help

	Error handling.

		Unexpected errors cause an immediate exit; this is an attempt to minimise error code:
			
			FLX_CTU_ASSERT						A developer error
			FLX_CTU_EXIT_UNEXPECTED_API_ERROR	Either an error specific to this machine, e.g.
												unable to read or write the trusted storage file,
												or a developer error, e.g. invalid handle.
		
		Errors that can reasonably be expected are reported explicitly. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "FlxActCommon.h"
#include "FlxActError.h"

#include "FlxActTransaction.h"

#include "compTranUtilContext.h"
#include "compTranUtilCommandDef.h"
#include "compTranUtilArgParser.h"
#include "compTranUtilCreateRequest.h"
#include "compTranUtilFulfillments.h"
#include "compTranUtilManageRequests.h"
#include "compTranUtilResponse.h"
#include "compTranUtilSend.h"

/* Private functions. */
static void sExit(FlxCtuContext* pContext, FlxCtuExitCode exitCode);
static const char* sGetExitCodeString(FlxCtuExitCode exitCode);
static void sPrintCommonLibraryInitError(int errorCode);

/***************************************************************************************************
*	Main entry point
*	================
*
*	argc	Number of command line arguments
*	argv	Array of command line arguments
*
* For usage see flxCtuHelpXXX functions in compTranUtilCommandDef.c
*
***************************************************************************************************/
int main(int argc, char * argv[])
{
	FlxActTranResult		tranResult;
	FlxActError				flxError;
	FlxActMode				mode = FLX_ACT_PREP;
	int						intResult = 0;

	const FlxCtuCommandDef* pCommandDefs = NULL; 
	size_t					commandDefCount = 0;
	FlxCtuContext*			pContext = NULL;
	FlxActBool				bSuccess = FLX_ACT_FALSE;
	const FlxCtuCommand*	pCommand = NULL;
	
	/* Get the command definitions - these define the arguments and options that can be used. */
	commandDefCount = flxCtuCommandDefsCreate(&pCommandDefs);

	/* Create a context to hold the commands parsed from the command line, status etc. */
	flxCtuContextCreate(pCommandDefs, commandDefCount, &pContext);

	/* Set the program name (argv[0]) in the context for use in messages. */
	FLX_CTU_ASSERT(argc > 0);
	pContext->pszProgramName = flxCtuStrDupNameFromPath(argv[0]);

	/* Parse the commands from the remaining command line arguments using the command definitions.
	   Creates a set of command objects within the context object. */
	if ( !flxCtuArgsParseToContext(pContext, (const char**)&argv[1], argc - 1))
		sExit(pContext, flxCtuExitInputError);

	/* Validate the commands - even though they conform to the definitions there are extra 
	   validation rules, for example -transaction requires either -new or -stored as well. */
	if ( !flxCtuContextValidateCommands(pContext) )
		sExit(pContext, flxCtuExitInputError);

	/* Process the "Brief" command (suppresses messages). */
	if (flxCtuContextHasCommand(pContext, CMD_BRIEF))
		flxCtuSetBrief(FLX_ACT_TRUE);

	/* Process the "Help" command - it is implicit if no commands at all are given. */
	pCommand = NULL;	/* In case no commands. */
	if ((pContext->commandCount == 0) || flxCtuContextGetCommandByName(pContext, CMD_HELP, &pCommand))
	{
		flxCtuCommandHelp(pContext, pCommand);
		sExit(pContext, flxCtuExitSuccess);
	}

	/* Initialise the Activation API and get a handle. This will fail if this executable has not been
	   processed with preptool since it was last built or if any FNP components are missing. */
	intResult = flxActCommonLibraryInit((const char*)0);
	if (intResult != 0)
	{
		sPrintCommonLibraryInitError(intResult);
		sExit(pContext, flxCtuExitInitError);
	}
	pContext->bCommonLibraryInitialised = FLX_ACT_TRUE;

	if ( !flxActCommonHandleOpen(&pContext->hApi, mode, &flxError) )
	{
		char logString[32];
		sprintf(logString, "%u", flxError.sysErrorNo, flxError.minorErrorNo);
		flxCtuPrintError("flxActCommonHandleOpen failed (%d,%d,%d)\n", 
						 flxError.majorErrorNo, 
						 flxError.minorErrorNo, 
						 flxError.sysErrorNo);
		sExit(pContext, flxCtuExitInitError);
	}
	/* Note server or application mode. */
	pContext->bIsServer	= (flxActCommonGetProtectionMode() == FLX_ACT_SVR)? 
						  FLX_ACT_TRUE : FLX_ACT_FALSE;

	/* Process commands that need an API handle but not a Transaction object. */
	/* Command  -view */
	if ( flxCtuContextGetCommandByName(pContext, CMD_VIEW, &pCommand) )
		bSuccess = flxCtuDoCommandView(pContext, pCommand);
	/* Command  -deleteFR */
	else if ( flxCtuContextGetCommandByName(pContext, CMD_DELETE_FR, &pCommand) )
		bSuccess = flxCtuDoCommandDeleteFR(pContext, pCommand);
	else
	{
		/* Create a transaction object - all the following commands require one. */
		bSuccess = flxActTransactionCreate(pContext->hApi, &pContext->hTransaction, &tranResult);
		if ( !bSuccess)
		{
			/* If reason for failure is TS load fail... */
			if (tranResult == FLX_ACT_TRAN_ERR_TS_LOAD)
			{
				/* It's probably because it is untrusted so do a local repair and try again. */
				(void)flxActCommonRepairLocalTrustedStorage(pContext->hApi);
				bSuccess = flxActTransactionCreate(pContext->hApi, &pContext->hTransaction, &tranResult);
			}
		}
		if ( !bSuccess)
		{
			flxCtuPrintError("flxActTransactionCreate failed %d\n", tranResult);
			sExit(pContext, flxCtuExitInitError);
		}

		/* Command  -transaction */
		if ( flxCtuContextGetCommandByName(pContext, CMD_TRANSACTION, &pCommand) )
			bSuccess = flxCtuDoCommandTransaction(pContext, pCommand);

		/* Command  -new */
		else if ( flxCtuContextGetCommandByName(pContext, CMD_NEW, &pCommand) )
			bSuccess = flxCtuDoCommandNew(pContext, pCommand);

		/* Command  -stored */
		else if ( flxCtuContextGetCommandByName(pContext, CMD_STORED, &pCommand) )
			bSuccess = flxCtuDoCommandStored(pContext, pCommand);

		/* Command  -process */
		else if ( flxCtuContextGetCommandByName(pContext, CMD_PROCESS, &pCommand) )
			bSuccess = flxCtuDoCommandProcess(pContext, pCommand);

		/* Command  -list */
		else if ( flxCtuContextGetCommandByName(pContext, CMD_LIST_REQUEST, &pCommand) )
			bSuccess = flxCtuDoCommandListRequests(pContext, pCommand);
		
		/* Command  -cancel */
		else if ( flxCtuContextGetCommandByName(pContext, CMD_CANCEL_REQUEST, &pCommand) )
			bSuccess = flxCtuDoCommandCancelRequest(pContext, pCommand);
		
		/* Command  -cancelAll */
		else if ( flxCtuContextGetCommandByName(pContext, CMD_CANCEL_ALL, &pCommand) )
			bSuccess = flxCtuDoCommandCancelAll(pContext, pCommand);

		else
		{
			flxCtuPrintError("Nothing to do.\n");
			sExit(pContext, flxCtuExitInputError);
		}
	}

	/* Exit. If there has been an error, the exit code will have been set in the context. */
	if (bSuccess)
		sExit(pContext, flxCtuExitSuccess);
	else
		sExit(pContext, pContext->exitCode);
}

/***************************************************************************************************
*	sExit 
*
* Clean up objects held in the context and the context itself, print the exit code and exit.
***************************************************************************************************/
static void sExit(FlxCtuContext* pContext, FlxCtuExitCode exitCode)
{
	if (pContext != NULL)
	{
		if (pContext->hTransaction != (FlxActTransaction)0)
			flxActTransactionDestroy(pContext->hTransaction);

		if (pContext->hApi != (FlxActHandle)0)
			flxActCommonHandleClose(pContext->hApi);

		if (pContext->bCommonLibraryInitialised)
			flxActCommonLibraryCleanup();

		flxCtuContextDestroy(pContext);
	}

	if (exitCode != flxCtuExitSuccess)
		flxCtuPrintMessage("Exit(%d) %s.\n", exitCode, sGetExitCodeString(exitCode));
	exit(exitCode);
}

/******************************************************************************
*	sGetExitCodeString
*
* Return a string describing an exit code.
******************************************************************************/
static const char* sGetExitCodeString(FlxCtuExitCode exitCode)
{
	switch (exitCode)
	{
		case flxCtuExitSuccess:					return "success";
		case flxCtuExitInputError:				return "error in command line";
		case flxCtuExitInitError:				return "FLEXnet initialisation error";
		case flxCtuExitCommsError:				return "communications error";
		case flxCtuExitResponseProcessingError:	return "error when processing response";
		case flxCtuExitApiError:				return "unexpected API failure";
		case flxCtuExitFileError:				return "file access error";
		case flxCtuExitMemoryError:				return "no memory";
		default:								return "";
	}				
}

/******************************************************************************
*	flxInitLoad error codes
******************************************************************************/

typedef enum FlxInitStatus
{
	flxInitSuccess 					= 0, 	/* Load was successful */
	flxInitUnableToLocate			= 1,	/* Unable to find security runtime */
	flxInitUnableToLoad				= 2,	/* Unable to load security runtime */
	flxInitUnsupportedPlatform		= 3,	/* OS version too old */
	flxInitInitializationError		= 4,	/* Initialization of security runtime failed */
	flxInitAllocationError			= 5,	/* Unable to allocate resources */
	flxInitNotProtected				= 6,	/* preptool has not been run since build */
	flxInitCannotReload				= 7,   	/* Call to flxInitLoad after an flxInitUnload */

	/* Mac OS X specific error values */
	flxInitEncodingError			= 8,	/* path string not in UTF-8? */
	flxInitUrlError					= 9,	/* error creating url */
	flxInitUrlNoPathError			= 10, 	/* error creating path from url */
	flxInitFrameworkNotLoadedError	= 11,	/* framework specified by bundle ID not loaded */
	flxInitBundleIDError			= 12,	/* Not a valid bundle ID */
	flxInitPathOverflow				= 13,	/* computed path is too long */

	/* Service error values */
	flxInitServiceNotInstalled		= 20,	/* service not installed */
	flxInitServiceNotEnoughRights	= 21	/* not enough rights to talk to service.
											Set service to start automatically to resolve this */
} FlxInitStatus;

/***************************************************************************************************
*	sPrintCommonLibraryInitError
*
* Print an error message, including text for the common causes of initialisation errors.
***************************************************************************************************/
static void sPrintCommonLibraryInitError(int errorCode)
{
	const char* pszErrorText = "";	/* Default. */

	switch (errorCode)
	{
	case flxInitUnableToLocate:				
						pszErrorText = "Unable to find FNP security runtime shared object";
						break;
	case flxInitNotProtected:			
						pszErrorText = "FNP preptool has not been run on this executable";
						break;
	case flxInitServiceNotInstalled:	
						pszErrorText = "FNP Service not installed on this machine.";
						break;
	case flxInitServiceNotEnoughRights:	
						pszErrorText = "Can't start FNP service (try setting automatic start)";
						break;
	}
	flxCtuPrintError("flxActCommonInit result %u   %s.\n", errorCode, pszErrorText);
}

/***************************************************************************************************
*	flxCtuOnNoMemory
*
* This will be called by the memory allocation functions on error.
* No messages to avoid using further resoiurcses.
***************************************************************************************************/
void flxCtuOnNoMemory()
{
	exit(flxCtuExitMemoryError);
}

/***************************************************************************************************
*	flxCtuOnUnexpectedApiError
*
* Target of unexpected fatal API errors - prints a message and exits.
***************************************************************************************************/
void flxCtuOnUnexpectedApiError(FlxCtuContext* pContext, const char* pszFilename, int line)
{
	flxCtuPrintError("Unexpected fatal API error in %s at line %u\n", pszFilename, line);
	flxCtuContextPrintLastApiError(pContext);											\
	sExit((FlxCtuContext*)pContext, flxCtuExitApiError);
}

/***************************************************************************************************
*	flxCtuOnAssert
*
* Target of failed asserts - prints a message and exits.
***************************************************************************************************/
void flxCtuOnAssert(int line, const char* pszFilename)
{
	flxCtuPrintError("Assert failed at line %u of file %s\n", line, pszFilename);
	exit(flxCtuExitAssert);
}
